"""
Utils for MQTT IO project.
"""
import asyncio
from typing import Any, Coroutine, List, Optional, cast


class PriorityCoro:
    """
    An object for adding a coroutine to an asyncio.PriorityQueue.
    """

    def __init__(self, coro: Coroutine[Any, Any, Any], priority: int):
        self.coro = coro
        self.priority = priority

    def __lt__(self, other: Any) -> bool:
        return cast(bool, self.priority < other.priority)

    def __eq__(self, other: Any) -> bool:
        return cast(bool, self.priority == other.priority)


def create_unawaited_task_threadsafe(
    loop: asyncio.AbstractEventLoop,
    transient_tasks: List["asyncio.Task[Any]"],
    coro: Coroutine[Any, Any, None],
    task_future: Optional["asyncio.Future[asyncio.Task[Any]]"] = None,
) -> None:
    """
    Schedule a coroutine on the loop and add the Task to transient_tasks.
    """

    def callback() -> None:
        task = loop.create_task(coro)
        transient_tasks.append(task)
        if task_future is not None:
            task_future.set_result(task)

    loop.call_soon_threadsafe(callback)



"""
Handles config validation and normalisation.
"""

import logging
from collections import Counter
from os.path import dirname, join, realpath
from typing import TYPE_CHECKING, Any, Dict, List, cast

import cerberus  # type: ignore
import yaml

from ..exceptions import ConfigValidationFailed
from ..types import ConfigType
from .validation.gpio import (
    validate_gpio_interrupt_for,
    validate_gpio_module_names,
    validate_gpio_modules_have_io_sections,
    validate_gpio_pins_only_configured_once,
)

if TYPE_CHECKING:
    from ..modules.gpio import GenericGPIO
    from ..modules.sensor import GenericSensor

_LOG = logging.getLogger(__name__)

# TODO: Tasks pending completion -@flyte at 02/03/2021, 10:37:38
# Add validation to make sure that the names used for inputs/outputs/stream modules
# are suitable for use in an MQTT topic.


class ConfigValidator(cerberus.Validator):  # type: ignore # No stubs for Cerberus
    """
    Cerberus Validator containing function(s) for use with validating or
    coercing values relevant to the MQTT IO project.
    """

    @staticmethod
    def _normalize_coerce_rstrip_slash(value: str) -> str:
        """
        Strip forward slashes from the end of the string.
        :param value: String to strip forward slashes from
        :type value: str
        :return: String without forward slashes on the end
        :rtype: str
        """
        return value.rstrip("/")

    @staticmethod
    def _normalize_coerce_tostring(value: Any) -> str:
        """
        Convert value to string.
        :param value: Value to convert
        :return: Value represented as a string.
        :rtype: str
        """
        return str(value)


def get_duplicate_names(io_conf: List[ConfigType]) -> List[str]:
    """
    Checks for any duplicated names in a list of dicts, such as the gpio_modules section
    of the config.
    """
    counter = Counter(x["name"] for x in io_conf)
    return [name for name, count in counter.items() if count > 1]


def get_main_schema() -> ConfigType:
    """
    Load the main config schema from the YAML file packaged in the same dir as this file.
    :return: Config schema
    :rtype: dict
    """
    schema_path = join(dirname(realpath(__file__)), "config.schema.yml")
    with open(schema_path) as schema_file:
        # We write this schema file, so we know it'll adhere to ConfigType rules
        return cast(ConfigType, yaml.safe_load(schema_file))


def get_main_schema_section(section: str) -> ConfigType:
    """
    Load the main config schema from the YAML file and return a specific section.
    """
    # We write this schema file, so we know it'll adhere to ConfigType rules.
    # Specifically, all top-level sections of the schema are also dicts.
    return cast(ConfigType, get_main_schema()[section]["schema"]["schema"])


def validate_and_normalise_config(
    config: Any, schema: Any, **validator_options: Any
) -> ConfigType:
    """
    Validate the config dict based on the config schema.
    :param config: The supplied configuration
    :type config: dict
    :param schema: The schema against which to validate the config
    :type schema: dict
    :return: Normalised config
    :rtype: dict
    """
    validator = ConfigValidator(schema, **validator_options)
    if not validator.validate(config):
        raise ConfigValidationFailed(
            "Config did not validate:\n%s" % yaml.dump(validator.errors)
        )
    validated_config: ConfigType = validator.normalized(config)
    return validated_config


def custom_validate_main_config(config: ConfigType) -> ConfigType:
    """
    Do our own validation on the config to make sure it contains sane, workable data.
    """
    unique_name_sections = (
        "gpio_modules",
        "sensor_modules",
        "stream_modules",
        "digital_inputs",
        "digital_outputs",
        "sensor_inputs",
        "stream_reads",
        "stream_writes",
    )

    # Make sure none of the lists use a name multiple times
    sections_with_duplicated_names = {}
    for config_section in unique_name_sections:
        duplicate_names = get_duplicate_names(config.get(config_section, []))
        if duplicate_names:
            sections_with_duplicated_names[config_section] = duplicate_names

    if sections_with_duplicated_names:
        raise ConfigValidationFailed(
            "Config did not validate due to duplicated names:\n%s"
            % yaml.dump(sections_with_duplicated_names)
        )

    bad_configs: Dict[str, Dict[str, List[str]]] = {}

    # Make sure each of the IO configs refer to an existing module config
    module_and_io_sections = dict(
        gpio_modules=("digital_inputs", "digital_outputs"),
        sensor_modules=("sensor_inputs",),
        stream_modules=("stream_reads", "stream_writes"),
    )
    for module_section, io_sections in module_and_io_sections.items():
        validate_gpio_module_names(bad_configs, config, module_section, io_sections)

    # Make sure all GPIO modules have entries in digital_inputs or digital_outputs
    validate_gpio_modules_have_io_sections(bad_configs, config)

    # Make sure all GPIO modules' pins are only configured once
    # validate_gpio_pins_only_configured_once(bad_configs, config)

    # Make sure all digital inputs listed in 'interrupt_for' lists are configured
    # as interrupts themselves.
    validate_gpio_interrupt_for(bad_configs, config["digital_inputs"])

    if bad_configs:
        raise ConfigValidationFailed(
            "Config did not validate due to errors in the following sections:\n%s"
            % yaml.dump(bad_configs)
        )
    return config


def load_main_config(path: str) -> ConfigType:
    """
    Read the main config file from the given path, validate it and return it normalised.
    :param path: The filesystem path
    :return: The config
    """
    with open(path, "r") as stream:
        raw_config = yaml.safe_load(stream)
    return validate_and_normalise_main_config(raw_config)


def validate_and_normalise_main_config(raw_config: Any) -> ConfigType:
    """
    Validate and normalise any raw config object with the main schema.
    """
    config = validate_and_normalise_config(raw_config, get_main_schema())
    config = custom_validate_main_config(config)
    return config


def validate_and_normalise_sensor_input_config(
    config: ConfigType, module: "GenericSensor"
) -> ConfigType:
    """
    Validate sensor input configs.
    """
    schema = get_main_schema()
    sensor_input_schema = schema["sensor_inputs"]["schema"]["schema"].copy()
    sensor_input_schema.update(getattr(module, "SENSOR_SCHEMA", {}))
    return validate_and_normalise_config(config, sensor_input_schema, allow_unknown=False)


def validate_and_normalise_digital_input_config(
    config: ConfigType, module: "GenericGPIO"
) -> ConfigType:
    """
    Validate digital input configs.
    """
    schema = get_main_schema()
    digital_input_schema = schema["digital_inputs"]["schema"]["schema"].copy()
    digital_input_schema.update(getattr(module, "PIN_SCHEMA", {}))
    digital_input_schema.update(getattr(module, "INPUT_SCHEMA", {}))
    return validate_and_normalise_config(
        config, digital_input_schema, allow_unknown=False
    )


def validate_and_normalise_digital_output_config(
    config: ConfigType, module: "GenericGPIO"
) -> ConfigType:
    """
    Validate digital output configs.
    """
    schema = get_main_schema()
    digital_output_schema = schema["digital_outputs"]["schema"]["schema"].copy()
    digital_output_schema.update(getattr(module, "PIN_SCHEMA", {}))
    digital_output_schema.update(getattr(module, "OUTPUT_SCHEMA", {}))
    return validate_and_normalise_config(
        config, digital_output_schema, allow_unknown=False
    )



"""
DS18S20/DS1822/DS18B20/DS1825/DS28EA00/MAX31850K temperature sensors
"""

from typing import Dict, cast

from ...types import ConfigType, SensorValueType
from . import GenericSensor

REQUIREMENTS = ("w1thermsensor>=2.0.0",)
ALLOWED_TYPES = ["DS18S20", "DS1822", "DS18B20", "DS1825", "DS28EA00", "MAX31850K"]
CONFIG_SCHEMA = {
    "address": dict(type="string", required=True, empty=False),
    "type": dict(
        type="string",
        required=True,
        empty=False,
        allowed=ALLOWED_TYPES + [x.lower() for x in ALLOWED_TYPES],
    ),
}


class Sensor(GenericSensor):
    """
    Implementation of Sensor class for the one wire temperature sensors. DS18B etc.
    """

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,import-error
        from w1thermsensor import W1ThermSensor  # type: ignore
        from w1thermsensor.sensors import Sensor as SensorType  # type: ignore

        sensor_types: Dict[str, SensorType] = {s.name: s for s in list(SensorType)}
        self.sensor_type = sensor_types[self.config["type"].upper()]
        self.sensor = W1ThermSensor(self.sensor_type, self.config["address"].lower())

    def get_value(self, sens_conf: ConfigType) -> SensorValueType:
        """
        Get the temperature value from the sensor
        """
        return cast(float, self.sensor.get_temperature())



"""
Event framework for subscribing to and firing events with asyncio callbacks.
"""

import asyncio
import logging
from abc import ABC
from dataclasses import dataclass
from typing import Any, Callable, Coroutine, Dict, List, Optional, Type

from .utils import create_unawaited_task_threadsafe

_LOG = logging.getLogger(__name__)

# Unfortunately can't use Callable[[Event], Coroutine[Any, Any, None]] here because
# Callable[] is 'contravariant'.
ListenerType = Callable[[Any], Coroutine[Any, Any, None]]


@dataclass
class Event(ABC):
    """
    Base Event class.
    """


@dataclass
class DigitalInputChangedEvent(Event):
    """
    A digital input changed state.
    """

    input_name: str
    from_value: Optional[bool]
    to_value: bool


@dataclass
class DigitalOutputChangedEvent(Event):
    """
    A digital output changed state.
    """

    output_name: str
    to_value: bool


@dataclass
class SensorReadEvent(Event):
    """
    A value was read from a sensor.
    """

    sensor_name: str
    value: Any


@dataclass
class StreamDataReadEvent(Event):
    """
    Some data was read from a stream.
    """

    stream_name: str
    data: bytes


@dataclass
class StreamDataSentEvent(Event):
    """
    Some data was written to a stream.
    """

    stream_name: str
    data: bytes


class EventBus:
    """
    Event bus that handles subscribing to specific events and firing coroutine callbacks.
    """

    def __init__(
        self,
        loop: asyncio.AbstractEventLoop,
        transient_tasks: List["asyncio.Task[Any]"],
    ):
        self._loop = loop
        self._transient_tasks = transient_tasks
        self._listeners: Dict[Type[Event], List[ListenerType]] = {}

    def fire(self, event: Event) -> List["asyncio.Future[asyncio.Task[Any]]"]:
        """
        Add callback functions that have subscribed to this event type to the task loop.
        """
        event_class = type(event)
        try:
            listeners = self._listeners[event_class]
            _LOG.debug(
                "Found %s listener(s) for event type %s",
                len(listeners),
                event_class.__name__,
            )
        except KeyError:
            _LOG.debug("No listeners for event type %s", event_class.__name__)
            return []

        task_futures: List["asyncio.Future[asyncio.Task[Any]]"] = []
        for listener in listeners:
            # Pass in a future on which the asyncio.Task will be set when the coro
            # has been scheduled on the loop.
            fut: "asyncio.Future[asyncio.Task[Any]]" = asyncio.Future()
            task_futures.append(fut)
            # Run threadsafe in case we're firing events from interrupt callback threads
            create_unawaited_task_threadsafe(
                self._loop, self._transient_tasks, listener(event), fut
            )
        return task_futures

    def subscribe(
        self,
        event_class: Type[Event],
        callback: ListenerType,
    ) -> Callable[[], None]:
        """
        Add a coroutine to be used as a callback when the given event class is fired.
        """
        if not isinstance(event_class, type):
            raise TypeError(
                "event_class must be of type 'type'. Got type %s." % type(event_class)
            )
        if not Event in event_class.mro():
            raise TypeError("Event class must be a subclass of mqtt_io.events.Event")
        if not callable(callback):
            raise TypeError("callback must be callable. Got type %s." % type(callback))
        self._listeners.setdefault(event_class, []).append(callback)

        def remove_listener() -> None:
            self._listeners[event_class].remove(callback)

        return remove_listener



"""
Contains the base class and some enums that are shared across all GPIO modules.
"""

# IDEA: Possible implementations -@flyte at 23/02/2021, 21:37:43
# Decide which executor we should use for async functions.
# https://docs.python.org/3/library/asyncio-eventloop.html#asyncio.loop.run_in_executor
# Should we use the default executor, or have one per module?

import abc
import asyncio
import logging
from concurrent.futures import ThreadPoolExecutor
from enum import Enum, Flag, auto
from typing import Any, Callable, Dict, Iterable, List, Optional

from ...types import ConfigType, PinType

_LOG = logging.getLogger(__name__)


class PinDirection(Enum):
    """
    Whether the GPIO pin is an input or an output.
    """

    INPUT = auto()
    OUTPUT = auto()


class PinPUD(Enum):
    """
    Whether the GPIO pin should be pulled up, down or not anywhere.
    """

    OFF = auto()
    UP = auto()
    DOWN = auto()


class InterruptEdge(Enum):
    """
    Whether to trigger an interrupt on rising edge, falling edge or both.
    """

    RISING = auto()
    FALLING = auto()
    BOTH = auto()


class InterruptSupport(Flag):
    """
    Classifies the kind of support a GPIO module has for interrupts.

    These are designed to be added to the GPIO module's `INTERRUPT_SUPPORT` class constant
    by ORing them together, for example:

    class GPIO(GenericGPIO):
        INTERRUPT_SUPPORT = (
            InterruptSupport.FLAG_REGISTER | InterruptSupport.CAPTURE_REGISTER
        )

    The `INTERRUPT_SUPPORT` constant will then be used when deciding how to handle
    interrupt triggered by, or for this GPIO module.
    """

    NONE = auto()

    # The library supporting this chip provides a function callback mechanism on interrupt
    SOFTWARE_CALLBACK = auto()

    # The chip flags which pin triggered the interrupt
    FLAG_REGISTER = auto()

    # The chip captures pin values at the moment of interrupt, so
    # we can read those to return the value that the pin changed to.
    CAPTURE_REGISTER = auto()

    # The chip changes the logic level on one or more dedicated
    # interrupt pins when one of its GPIO pin's logic level changes.
    INTERRUPT_PIN = auto()

    # The chip allows you to set which pins trigger interrupts
    SET_TRIGGERS = auto()


class GenericGPIO(abc.ABC):
    """
    Abstracts a generic GPIO interface to be implemented by the modules in this
    directory.
    """

    INTERRUPT_SUPPORT = InterruptSupport.NONE

    def __init__(self, config: ConfigType):
        self.config = config
        self.pin_configs: Dict[PinType, ConfigType] = {}
        self.io: Any = None  # pylint: disable=invalid-name
        self.interrupt_edges: Dict[PinType, InterruptEdge] = {}

        self.direction_map: Dict[PinDirection, Any] = {}
        self.pullup_map: Dict[PinPUD, Any] = {}
        self.interrupt_edge_map: Dict[InterruptEdge, Any] = {}

        self.setup_module()

    @abc.abstractmethod
    def setup_module(self) -> None:
        """
        Called on initialisation of the GPIO module during the startup phase.

        The module's config from the `gpio_modules` section of the config file is stored
        in `self.config`.
        """

    @abc.abstractmethod
    def setup_pin(
        self,
        pin: PinType,
        direction: PinDirection,
        pullup: PinPUD,
        pin_config: ConfigType,
        initial: Optional[str] = None,
    ) -> None:
        """
        Called on initialisation of each pin of the GPIO module during the startup phase.

        The `pin_config` passed in here is the pin's entry in the `digital_inputs` or
        `digital_outputs` section of the config file.
        """

    @abc.abstractmethod
    def set_pin(self, pin: PinType, value: bool) -> None:
        """
        Set an individual pin to the given value.
        """

    @abc.abstractmethod
    def get_pin(self, pin: PinType) -> bool:
        """
        Poll a pin for its value.
        """

    def setup_interrupt(
        self,
        pin: PinType,
        edge: InterruptEdge,
        in_conf: ConfigType,
    ) -> None:
        """
        Configure a pin as an interrupt.

        This is used on modules which don't supply software callbacks, but use some other
        way of representing interrupts, such as connecting a dedicated interrupt output
        pin to another module that does supply software callbacks.
        """

    def setup_interrupt_callback(
        self,
        pin: PinType,
        edge: InterruptEdge,
        in_conf: ConfigType,
        callback: Callable[..., None],
    ) -> None:
        """
        Configure a pin as an interrupt and set up the callback function to be called when
        the interrupt is triggered.
        """

    def setup_interrupt_internal(
        self,
        pin: PinType,
        edge: InterruptEdge,
        in_conf: ConfigType,
        callback: Optional[Callable[[List[Any], Dict[Any, Any]], None]] = None,
    ) -> None:
        """
        Used internally to ensure that `self.interrupt_edges` is updated for this pin.
        """
        self.interrupt_edges[pin] = edge
        if callback is None:
            return self.setup_interrupt(pin, edge, in_conf)
        return self.setup_interrupt_callback(pin, edge, in_conf, callback)

    def cleanup(self) -> None:
        """
        Called when closing the program to handle any cleanup operations.
        """

    def setup_pin_internal(self, direction: PinDirection, pin_config: ConfigType) -> None:
        """
        Called internally to setup a pin just by supplying config. Calls setup_pin()
        that's been implemented by the GPIO module.
        """
        pin: PinType = pin_config["pin"]
        self.pin_configs[pin] = pin_config
        pud = PinPUD.OFF
        if pin_config.get("pullup"):
            pud = PinPUD.UP
        elif pin_config.get("pulldown"):
            pud = PinPUD.DOWN
        for key in ("direction", "pullup", "pulldown"):
            try:
                del pin_config[key]
            except KeyError:
                continue
        return self.setup_pin(
            pin, direction, pud, pin_config=pin_config, initial=pin_config.get("initial")
        )

    def remote_interrupt_for(self, pin: PinType) -> List[str]:
        """
        Return the list of pin names that this pin is a remote interrupt for.
        """
        return self.pin_configs[pin].get("interrupt_for", [])

    def get_int_pins(self) -> List[PinType]:
        """
        If the chip has a register that logs which pins caused the last interrupt,
        then we read it here and return the list of pins.
        """
        raise NotImplementedError()

    async def async_get_int_pins(self) -> List[PinType]:
        """
        Use a ThreadPoolExecutor to call the module's synchronous set_pin function.
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(ThreadPoolExecutor(), self.get_int_pins)

    def get_captured_int_pin_values(
        self, pins: Optional[Iterable[PinType]] = None
    ) -> Dict[PinType, bool]:
        """
        If the chip has a register that logs the values of the pins at the point of
        the last interrupt, then we read it here and return a dict.

        If pins is None, then we get the whole register and return it, otherwise
        just return the values for the pins requested.
        """
        raise NotImplementedError()

    async def async_get_captured_int_pin_values(
        self, pins: Optional[Iterable[PinType]] = None
    ) -> Dict[PinType, bool]:
        """
        Use a ThreadPoolExecutor to call the module's synchronous
        get_captured_int_pin_values function.
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(
            ThreadPoolExecutor(), self.get_captured_int_pin_values, pins
        )

    async def async_set_pin(self, pin: PinType, value: bool) -> None:
        """
        Use a ThreadPoolExecutor to call the module's synchronous set_pin function.
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(ThreadPoolExecutor(), self.set_pin, pin, value)

    async def async_get_pin(self, pin: PinType) -> bool:
        """
        Use a ThreadPoolExecutor to call the module's synchronous get_pin function.
        """
        loop = asyncio.get_event_loop()
        return await loop.run_in_executor(ThreadPoolExecutor(), self.get_pin, pin)

    def get_interrupt_value(self, pin: PinType, *args: Any, **kwargs: Any) -> bool:
        """
        Called on interrupt when this module's software callback was called, in order to
        get a value for the pin that caused the interrupt. This could just involve polling
        for the value, or looking it up in a register that captures the value at time of
        interrupt.
        """

    async def get_interrupt_values_remote(
        self, pins: List[PinType]
    ) -> Dict[PinType, bool]:
        """
        Called when another module's input pin was triggered by our module's interrupt
        pin's output changing. Takes a list of pins that can potentially trigger this
        interrupt, as defined in the `interrupt_for` list in the config file.
        """
        pins_set = set(pins)
        if self.INTERRUPT_SUPPORT & InterruptSupport.FLAG_REGISTER:
            int_pins = set(await self.async_get_int_pins())
            matching_pins = pins_set.intersection(int_pins)
            if not matching_pins:
                _LOG.warning(
                    (
                        "An interrupt was triggered for %s but none of the interrupting "
                        "pins returned by the GPIO device (%s) matched the pins "
                        "configured in `interrupt_for` (%s)."
                    ),
                    self,
                    int_pins,
                    pins,
                )
                return {}
        else:
            matching_pins = pins_set

        pin_values: Dict[PinType, bool]
        if self.INTERRUPT_SUPPORT & InterruptSupport.CAPTURE_REGISTER:
            pin_values = await self.async_get_captured_int_pin_values(pins=matching_pins)
        else:
            pin_values = {}
            for pin in matching_pins:
                try:
                    edge = self.interrupt_edges[pin]
                except KeyError:
                    _LOG.error(
                        (
                            "Pin %s on %s did not have an entry in the `interrupt_edges` "
                            "dict. This means that the pin probably wasn't configured "
                            "with an `interrupt` value in its config."
                        ),
                        pin,
                        self,
                    )
                    continue
                if edge == InterruptEdge.BOTH:
                    # We'll have to poll the pin for its current value
                    pin_values[pin] = await self.async_get_pin(pin)
                else:
                    # NOTE: Needs discussion or investigation -@flyte at 18/02/2021, 10:37:20
                    # This might cause false interrupts on chips where there's multiple
                    # pins configured as interrupts on a single remote interrupt line.
                    # If pin 0 changes value but pin 1 doesn't, and the chip doesn't
                    # provide a flag register, then this might just report that pin 1
                    # changed to its interrupt state (high or low for rising and falling
                    # respectively) when it did not, and it was just pin 0 that changed.
                    #
                    # It's probably best to make this configurable, because the
                    # alternative is just to poll the pin every time, but that could cause
                    # interrupts to report wrong values for the pin if we poll it after
                    # it's changed back to the opposite logic level. :thinking:

                    # We can infer the value based on how it's configured
                    pin_values[pin] = edge == InterruptEdge.RISING

        return pin_values



"""
Linux Kernel 4.8+ libgpiod
"""
import threading
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional

from ...types import ConfigType, PinType
from . import GenericGPIO, InterruptEdge, InterruptSupport, PinDirection, PinPUD

if TYPE_CHECKING:
    # pylint: disable=import-error
    import gpiod  # type: ignore

# Requires libgpiod-devel, libgpiod
REQUIREMENTS = ("gpiod",)

CONFIG_SCHEMA = {
    "chip": {"type": "string", "required": False, "default": "/dev/gpiochip0"}
}


class GPIO(GenericGPIO):
    """
    Implementation of GPIO class for libgpiod (linux kernel >= 4.8).
    """

    INTERRUPT_SUPPORT = InterruptSupport.SOFTWARE_CALLBACK

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,import-error
        import gpiod

        self.io: gpiod = gpiod
        self.chip = gpiod.chip(self.config["chip"])
        self.pins: Dict[PinType, gpiod.line] = {}
        self.interrupt_threads: Dict[PinType, threading.Thread] = {}
        self.stop_event = threading.Event()

        self.direction_map = {
            PinDirection.INPUT: gpiod.line_request.DIRECTION_INPUT,
            PinDirection.OUTPUT: gpiod.line_request.DIRECTION_OUTPUT,
        }

        self.interrupt_edge_map = {
            InterruptEdge.RISING: gpiod.line_request.EVENT_RISING_EDGE,
            InterruptEdge.FALLING: gpiod.line_request.EVENT_FALLING_EDGE,
            InterruptEdge.BOTH: gpiod.line_request.EVENT_BOTH_EDGES,
        }

    def setup_pin(
        self,
        pin: PinType,
        direction: PinDirection,
        pullup: PinPUD,
        pin_config: ConfigType,
        initial: Optional[str] = None,
    ) -> None:
        """
        Setup a pin as either input or output.
        pin:        offset to use the gpio line
        direction:  input or output
        pullup:     pullup settings are not supported
        """
        # Pullup settings are called bias in libgpiod and are only
        # available since Linux Kernel 5.5. They are as of now not
        # yet part of python3-gpiod.

        line: "gpiod.line" = self.chip.get_line(pin)

        line_request = self.io.line_request()
        line_request.consumer = "mqtt-io"
        line_request.request_type = self.direction_map[direction]

        line.request(line_request)
        if direction == PinDirection.OUTPUT and initial is not None:
            line.set_value(1 if initial == "high" else 0)
        self.pins[pin] = line

    def setup_interrupt_callback(
        self,
        pin: PinType,
        edge: InterruptEdge,
        in_conf: ConfigType,
        callback: Callable[..., None],
    ) -> None:
        """
        Install interrupt callback function
        handle:     is returned in the callback function as identification
        pin:        gpio to watch for interrupts
        edge:       triggering edge: RISING, FALLING or BOTH
        callback:   the callback function to be called, when interrupt occurs
        bouncetime: minimum time between two interrupts
        """

        line_request = self.io.line_request()
        line_request.consumer = "mqtt-io"
        line_request.request_type = self.interrupt_edge_map[edge]

        bouncetime: int = in_conf["bouncetime"]
        int_thread = InterruptThread(
            self.chip, pin, line_request, callback, bouncetime, self.stop_event
        )
        int_thread.start()
        self.interrupt_threads[pin] = int_thread

    def get_interrupt_value(self, pin: PinType, *args: Any, **kwargs: Any) -> bool:
        # We established the pin's value in the InterruptThread, so we just give it back
        pin_value: bool = kwargs["pin_value"]
        return pin_value

    def set_pin(self, pin: PinType, value: bool) -> None:
        self.pins[pin].set_value(value)

    def get_pin(self, pin: PinType) -> bool:
        return bool(self.pins[pin].get_value())

    def cleanup(self) -> None:
        self.stop_event.set()
        for thread in self.interrupt_threads.values():
            thread.join(timeout=10)


class InterruptThread(threading.Thread):
    """
    Thread that waits on interrupt events for a given pin, then calls the callback.
    """

    def __init__(
        self,
        chip: "gpiod.chip",
        pin: PinType,
        line_request: "gpiod.line_request",
        callback: Callable[..., None],
        bouncetime: int,
        stop_event: threading.Event,
    ):
        super().__init__()
        self.daemon = True

        self.pin = pin
        self.line: "gpiod.line" = chip.get_line(self.pin)
        self.line.release()
        self.line.request(line_request)
        self.callback = callback
        self.bouncetime = timedelta(milliseconds=bouncetime)
        self.stop_event = stop_event

    def run(self) -> None:
        # pylint: disable=import-outside-toplevel,import-error
        import gpiod

        previous_event_time = datetime.now()
        while not self.stop_event.is_set():
            if not self.line.event_wait(timedelta(seconds=2)):
                continue
            event: gpiod.line_event = self.line.event_read()
            now = datetime.now()
            if now - previous_event_time < self.bouncetime:
                continue
            previous_event_time = now
            pin_value = None
            if event.event_type == gpiod.line_request.EVENT_RISING_EDGE:
                pin_value = True
            elif event.event_type == gpiod.line_request.EVENT_FALLING_EDGE:
                pin_value = False
            if pin_value is None:
                # Poll the pin for its value :(
                pin_value = bool(self.line.get_value())
            self.callback(pin_value=pin_value)



import asyncio
from typing import Any, Dict, Type

import yaml
from behave import given, then, when  # type: ignore
from behave.api.async_step import async_run_until_complete  # type: ignore
from mqtt_io import events
from mqtt_io.server import MqttIo

try:
    from unittest.mock import AsyncMock  # type: ignore[attr-defined]
except ImportError:
    from mock import AsyncMock  # type: ignore[attr-defined]

# pylint: disable=function-redefined,protected-access


@when("we subscribe to {event_type_name}")
def step(context: Any, event_type_name: str) -> None:
    mqttio: MqttIo = context.data["mqttio"]
    event_type = getattr(events, event_type_name)
    mock_sub = AsyncMock()
    context.data["event_subs"][event_type_name] = mock_sub
    mqttio.event_bus.subscribe(event_type, mock_sub)


@when("we fire a new {event_type_name} event with")  # type: ignore[no-redef]
@async_run_until_complete(loop="loop")
async def step(context: Any, event_type_name: str) -> None:
    data: Dict[str, Any] = yaml.safe_load(context.text)
    assert isinstance(data, dict), "Data provided to this step must be a YAML dict"
    mqttio: MqttIo = context.data["mqttio"]
    event_type: Type[events.Event] = getattr(events, event_type_name)
    event = event_type(**data)
    task_futures = mqttio.event_bus.fire(event)
    await asyncio.gather(*task_futures)


@then("{func_name} function should be subscribed to {event_type_name}")  # type: ignore[no-redef]
def step(context: Any, func_name: str, event_type_name: str) -> None:
    mqttio = context.data["mqttio"]
    event_type = getattr(events, event_type_name)
    assert func_name in {x.__name__ for x in mqttio.event_bus._listeners[event_type]}


@then("{event_type_name} is fired with")  # type: ignore[no-redef]
def step(context: Any, event_type_name: str) -> None:
    data = yaml.safe_load(context.text)
    assert isinstance(data, dict), "Data provided to this step must be a YAML dict"
    # event_type = getattr(events, event_type_name)
    mock_sub = context.data["event_subs"][event_type_name]
    event = mock_sub.call_args.args[0]
    for key, value in data.items():
        assert getattr(event, key) == value, f"Expecting event.{key} to be {value}"



"""
Shared types and utils for validation functions.
"""

from typing import Dict, List


BadConfigsType = Dict[str, Dict[str, List[str]]]


def add_error(
    bad_configs: BadConfigsType,
    section: str,
    subsection: str,
    error: str,
) -> None:
    """
    Add an error to the given section and subsection of bad_configs.
    """
    bad_configs.setdefault(section, {}).setdefault(subsection, []).append(error)



"""
Orange Pi GPIO
"""

from typing import Optional

from ...types import ConfigType, PinType
from . import GenericGPIO, PinDirection, PinPUD

ALLOWED_BOARDS = [
    "zero",
    "r1",
    "zeroplus",
    "zeroplus2h5",
    "zeroplus2h3",
    "pcpcplus",
    "one",
    "lite",
    "plus2e",
    "pc2",
    "prime",
]
ALLOWED_MODES = ["bcm", "board", "mode_soc"]
REQUIREMENTS = ("OrangePi.GPIO",)
CONFIG_SCHEMA = {
    "board": {
        "type": "string",
        "required": True,
        "empty": False,
        "allowed": ALLOWED_BOARDS + list(map(str.upper, ALLOWED_BOARDS)),
    },
    "mode": {
        "type": "string",
        "required": True,
        "empty": False,
        "default": "bcm",
        "allowed": ALLOWED_MODES + list(map(str.upper, ALLOWED_MODES)),
    },
}


class GPIO(GenericGPIO):
    """
    Implementation of GPIO class for Orange Pi native GPIO.
    """

    def setup_module(self) -> None:
        # pylint: disable=import-outside-toplevel,import-error
        import OPi.GPIO as gpio  # type: ignore

        self.io = gpio
        self.direction_map = {
            PinDirection.INPUT: gpio.IN,
            PinDirection.OUTPUT: gpio.OUT,
        }
        self.pullup_map = {
            PinPUD.OFF: gpio.PUD_OFF,
            PinPUD.UP: gpio.PUD_UP,
            PinPUD.DOWN: gpio.PUD_DOWN,
        }

        board = self.config["board"].upper()
        mode = self.config["mode"].upper()
        if not hasattr(gpio, board):
            raise AssertionError("%s board not found" % board)
        gpio.setboard(getattr(gpio, board))
        gpio.setmode(getattr(gpio, mode))

    def setup_pin(
        self,
        pin: PinType,
        direction: PinDirection,
        pullup: PinPUD,
        pin_config: ConfigType,
        initial: Optional[str] = None,
    ) -> None:
        direction = self.direction_map[direction]

        if pullup is None:
            pullup = self.pullup_map[PinPUD.OFF]
        else:
            pullup = self.pullup_map[pullup]

        initial_int = {None: -1, "low": 0, "high": 1}[initial]
        try:
            self.io.setup(pin, direction, pull_up_down=pullup, initial=initial_int)
        except ValueError as exc:
            raise IOError(f"channel {pin} setup failed") from exc

    def set_pin(self, pin: PinType, value: bool) -> None:
        self.io.output(pin, value)

    def get_pin(self, pin: PinType) -> bool:
        return bool(self.io.input(pin))

    def cleanup(self) -> None:
        self.io.cleanup()



"""
Mock GPIO module for using with the tests.
"""

from typing import Any, Callable, Dict, Iterable, List, Optional
from unittest.mock import Mock

from ...types import ConfigType, PinType
from . import GenericGPIO, InterruptEdge, InterruptSupport, PinDirection, PinPUD

REQUIREMENTS = ()
CONFIG_SCHEMA = dict(test=dict(type="boolean", required=False, default=False))


# pylint: disable=useless-super-delegation,too-many-instance-attributes
class GPIO(GenericGPIO):
    """
    Mock GPIO module for using with the tests.
    """

    INTERRUPT_SUPPORT = (
        InterruptSupport.SOFTWARE_CALLBACK
        | InterruptSupport.FLAG_REGISTER
        | InterruptSupport.CAPTURE_REGISTER
    )
    PIN_SCHEMA = {
        "test": dict(type="boolean", required=False, default=False),
    }

    def __init__(self, config: ConfigType):
        self.setup_module = Mock()  # type: ignore[assignment]
        self.setup_pin = Mock()  # type: ignore[assignment]
        self.setup_interrupt = Mock()  # type: ignore[assignment]
        self.setup_interrupt_callback = Mock()  # type: ignore[assignment]
        self.set_pin = Mock()  # type: ignore[assignment]
        self.get_pin = Mock(return_value=True)  # type: ignore[assignment]
        self.get_int_pins = Mock(return_value=1)  # type: ignore[assignment]
        self.get_captured_int_pin_values = Mock(return_value={1: 1})  # type: ignore[assignment]

        super().__init__(config)
        self.interrupt_callbacks: Dict[
            PinType, Callable[[List[Any], Dict[Any, Any]], None]
        ] = {}

    def setup_interrupt_callback(
        self,
        pin: PinType,
        edge: InterruptEdge,
        in_conf: ConfigType,
        callback: Callable[[List[Any], Dict[Any, Any]], None],
    ) -> None:
        self.interrupt_callbacks[pin] = callback

    def setup_module(self) -> None:
        return super().setup_module()

    def setup_pin(
        self,
        pin: PinType,
        direction: PinDirection,
        pullup: PinPUD,
        pin_config: ConfigType,
        initial: Optional[str] = None,
    ) -> None:
        return super().setup_pin(pin, direction, pullup, pin_config, initial=initial)

    def setup_interrupt(
        self, pin: PinType, edge: InterruptEdge, in_conf: ConfigType
    ) -> None:
        return super().setup_interrupt(pin, edge, in_conf)

    def set_pin(self, pin: PinType, value: bool) -> None:
        return super().set_pin(pin, value)

    def get_pin(self, pin: PinType) -> bool:
        return super().get_pin(pin)

    def get_int_pins(self) -> List[PinType]:
        return super().get_int_pins()

    def get_captured_int_pin_values(
        self, pins: Optional[Iterable[PinType]] = None
    ) -> Dict[PinType, bool]:
        return super().get_captured_int_pin_values(pins=pins)



"""
Validation functions for GPIO configs.
"""

from collections import Counter
from typing import Any, Dict, Iterable, List, Set

from ...types import ConfigType
from . import BadConfigsType, add_error


def validate_gpio_module_names(
    bad_configs: BadConfigsType,
    config: ConfigType,
    module_section: str,
    io_sections: Iterable[str],
) -> None:
    """
    Ensure that all of the IO config sections refer to existing module names.
    """
    module_names = {x["name"] for x in config.get(module_section, [])}

    for io_section in io_sections:
        for io_conf in config.get(io_section, []):
            if io_conf["module"] not in module_names:
                add_error(
                    bad_configs,
                    io_section,
                    io_conf["name"],
                    "%s section has no module configured with the name '%s'"
                    % (
                        module_section,
                        io_conf["module"],
                    ),
                )


def validate_gpio_modules_have_io_sections(
    bad_configs: BadConfigsType,
    config: ConfigType,
) -> None:
    """
    Ensure that all GPIO modules have an IO section.
    """
    digital_inputs = config.get("digital_inputs", [])
    digital_outputs = config.get("digital_outputs", [])
    modules_used = set(x["module"] for x in digital_inputs + digital_outputs)

    for module in config.get("gpio_modules", []):
        if module["name"] not in modules_used:
            add_error(
                bad_configs,
                "gpio_modules",
                module["name"],
                "GPIO module '%s' does not have any digital_input or digital_output configs"
                % module["name"],
            )


def validate_gpio_pins_only_configured_once(
    bad_configs: BadConfigsType,
    config: ConfigType,
) -> None:
    """
    Ensure that any GPIO pins are only configured once.
    """
    digital_inputs: List[Dict[str, Any]] = config.get("digital_inputs", [])
    digital_outputs: List[Dict[str, Any]] = config.get("digital_outputs", [])
    io_confs_per_module: Dict[str, List[Dict[str, Any]]] = {}
    for io_conf in digital_inputs + digital_outputs:
        io_confs_per_module.setdefault(io_conf["module"], []).append(io_conf)
    for module_name, io_confs in io_confs_per_module.items():
        pin_config_counts = Counter(io_conf["pin"] for io_conf in io_confs)
        duplicated_pins = [pin for pin, count in pin_config_counts.items() if count > 1]
        for dup_pin in duplicated_pins:
            add_error(
                bad_configs,
                "gpio_modules",
                module_name,
                "Pin '%s' has more than one configuration in digital_inputs or digital_outputs"
                % dup_pin,
            )


def validate_gpio_interrupt_for(
    bad_configs: BadConfigsType,
    digital_inputs: List[ConfigType],
) -> None:
    """
    Ensure that:
    - Pins listed in interrupt_for are configured as interrupts
    - A pin's interrupt_for doesn't contain itself
    - A pin with interrupt_for is configured as an interrupt
    """
    interrupt_pins: Set[str] = set(
        in_conf["name"] for in_conf in digital_inputs if in_conf.get("interrupt")
    )
    for in_conf in [x for x in digital_inputs if x.get("interrupt_for")]:
        errors = []
        if "interrupt" not in in_conf:
            errors.append(
                "Pin is configured as an 'interrupt_for' other pins, but is not "
                "configured as an interrupt itself"
            )
        for pin_name in in_conf["interrupt_for"]:
            if pin_name not in interrupt_pins:
                errors.append(
                    f"Pin '{pin_name}' listed in 'interrupt_for' is not configured as an "
                    "interrupt itself"
                )
            if pin_name == in_conf["name"]:
                errors.append(f"Pin '{pin_name}' lists itself in 'interrupt_for'")

        for error in errors:
            add_error(bad_configs, "digital_inputs", in_conf["name"], error)
